
"""
# /**********************************************************************************
#* Copyright (C) 2017-2020 by KEYARK, INC. - All Rights Reserved.
#* All information contained herein is proprietary and confidential to KEYARK, INC.
#* Any use, reproduction, or disclosure without the written
#* permission of KEYARK, INC. is prohibited.
#* Written by Natalia Belyaeva <nbelyaeva@keyark.com>
#***********************************************************************************/
"""

import locale
locale.setlocale(locale.LC_ALL, 'en_US.utf-8')
import time
from typing import Optional

from keysql_client.keysql_connector import *
from keysql_client.keysql_statement import KeySqlStatement
from keysql_client.keysql_show_statement import KeySqlShowStatement
from keysql_client.keysql_exception import KeySqlExceptionMessage, KeySqlException
from keysql_client.keysql_reply_statement import KeySqlReplyStatement
from keysql_client.keysql_parser_query_parameterized import KeySqlParserQueryParameterized
from keysql_client.keysql_parser_packed_typed_format import KeySqlParserPackedFormat

from typing import List
import json
import os
from collections import namedtuple
from json import JSONEncoder
from keysql_client.keysql_parser import KeySqlParser

import logging
import signal
import time
import errno
import functools

class TimeoutError(Exception):
    pass

#@staticmethod
class KeySqlConnectionConstants:
    """Connection with KeySql server
    """
    QUERY_TIMEOUT_IMPLEMENTED = os.name == 'posix'
    QUERY_TIMEOUT_SEC = -2   # -1 - unlimited, example: 20 sec
    CONNECT_TIMEOUT_IMPLEMENTED = os.name == 'posix'
    CONNECT_TIMEOUT_SEC = 20 # -1 - if unlimited assigned , example: 5 sec

def timeout_server_connect(seconds=KeySqlConnectionConstants.QUERY_TIMEOUT_SEC, error_message=os.strerror(errno.ETIME)):
    def decorator(func):
        def _handle_timeout(signum, frame):
            print("=== server_timeout_sec: ", KeySqlConnectionConstants.CONNECT_TIMEOUT_SEC, "===")
            raise TimeoutError(error_message)
        if (not KeySqlConnectionConstants.CONNECT_TIMEOUT_IMPLEMENTED) or (KeySqlConnectionConstants.CONNECT_TIMEOUT_SEC < 0):
            return func # decorator

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            signal.signal(signal.SIGALRM, _handle_timeout)
            signal.alarm(seconds)
            try:
                result = func(*args, **kwargs)
            finally:
                signal.alarm(0)
            return result
        return wrapper
    return decorator

def timeout(seconds=KeySqlConnectionConstants.QUERY_TIMEOUT_SEC, error_message=os.strerror(errno.ETIME)):
    def decorator(func):
        def _handle_timeout(signum, frame):
            print("=== self.query_timeout_sec: ", KeySqlConnectionConstants.QUERY_TIMEOUT_SEC, "===")
            raise TimeoutError(error_message)

        if not KeySqlConnectionConstants.QUERY_TIMEOUT_IMPLEMENTED or KeySqlConnectionConstants.QUERY_TIMEOUT_SEC < 0:
            return func # decorator

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            signal.signal(signal.SIGALRM, _handle_timeout)
            signal.alarm(seconds)
            try:
                result = func(*args, **kwargs)
            finally:
                signal.alarm(0)
            return result
        return wrapper
    return decorator

class KeySqlConnection:
    """Connection with KeySql server
    """
    last_str_result = ""
    schema = ""
    owner = None
    cum_execute_time_in_sec = 0.0

    #QUERY_TIMEOUT_IMPLEMENTED = os.name == 'posix'
    #QUERY_TIMEOUT_SEC = 32600 # not assigned - 1 day , example: 20 sec
    #CONNECT_TIMEOUT_IMPLEMENTED = os.name == 'posix'
    #CONNECT_TIMEOUT_SEC = 32600 # not assigned , example: 5 sec

    @timeout_server_connect(KeySqlConnectionConstants.CONNECT_TIMEOUT_SEC)
    def __init__(self, host : str, port : int, user: str, password : str,
                 ssl : bool = True, schema: str = "", owner: Optional[str] = None,
                 query_timeout_sec = KeySqlConnectionConstants.QUERY_TIMEOUT_SEC,
                 server_timeout_sec = KeySqlConnectionConstants.CONNECT_TIMEOUT_SEC):  # , nonblocking = True):
        self._connector = KeySqlConnector(host, port, schema)
        #print("user, password, schema:", user, password, schema, '================================================')
        self.schema = schema
        self.owner = owner # None
        self.user_name = user
        self._reply_header = None
        self.ssl = ssl
        self.query_timeout_sec = query_timeout_sec
        self.server_timeout_sec = server_timeout_sec
        KeySqlConnectionConstants.QUERY_TIMEOUT_SEC = self.query_timeout_sec
        KeySqlConnectionConstants.CONNECT_TIMEOUT_SEC = self.server_timeout_sec
        try:
            if ssl:
                self._connector.ssl_connect(user,password, schema)  # , nonblocking=nonblocking)
            else:
                self._connector._connect(user, password, schema)    #, nonblocking=nonblocking)
        except TimeoutError as ex:
            logging.exception(KeySqlExceptionMessage.SERVER_TIMEOUT_EXCEPTION)
            print(ex, 'connection timeout happened ==================================')
            raise ex


    def __exit__(self):
        self._connector._disconnect()


    def disconnect(self):
        """Disconnect with KeySql server
        :return: None
        """
        self.__exit__()

    def shutdown(self):
        if self.ssl:
            raise KeySqlConnectException(KeySqlExceptionMessage.SHUTDOWN_TCP_ONLY)
        else:
            self._connector.shutdown()

    def connector(self):
        return self._connector

    def is_connected(self):
        """Check if connected with KeySql server
        :return: True or False
        """
        return self._connector.is_connected()

    def change_schema(self, schema, owner=None) -> bool:
        """Check: if schema is the same as current schema, do nothing
           Otherwise: change schema
        :return: True (schema changed) or False (schema unchanged)
        """
        if self.schema == schema and owner is None:
            return False
        try:
            changed = self._connector.reconnect(schema)
            if changed:
                self.schema = schema
            self.owner = owner

            return changed
        except KeySqlException as err: # (-1, KeySqlExceptionMessage.CAN_NOT_CHANGE_SCHEMA) as err:
            short_message = self._connector.cut_message(str(err.message))
            logging.exception(short_message)
            raise

    def get_schema(self):
        return self.schema

    def get_user_name(self):
        return self.user_name


    def get_header_reply(self):
        """Get header
        :return: Dictionary of attributes, header['exit_code'] = 0 means successful query
        """
        return self._reply_header


    def get_last_str_result(self):
        return self.last_str_result


    def customKeySqlCatalogDecoder(catalogDict):
        return namedtuple('X', catalogDict.keys())(*catalogDict.values())

    # format_output: 'keysql', 'json'

    @timeout(KeySqlConnectionConstants.QUERY_TIMEOUT_SEC)
    def execute(self, statement, *args, base_classes: List=[], keep_str_res=False, check_inj=False, format='keysql', format_out='keysql'):
        """Execute query on KeySql server
        :param statement (string, list of class:
        :param base_classes: List of classes to map (default: []  (not used now))
        :param keep_str_res  =  keep string response from server (how it is shown in KeySql studio)
                 To retrive string response:  self.get_last_str_result().   Default_value=False,
        :param check_inj=False,  - not used now
        :param format   - format of string to get from server (for get_last_str_result() function). Default_value='keysql'
        :param format_out - in what format to output result ('class"||json). Default_value='keysqclass' (python class).
                As of now, format_out = format
        :return: a list of classes||json    For one element the result is list of one element [elem]
        :raise: Exception if query can not be performed (no connection, wrong syntax, etc.)
        """

        #logging.info('start query: ' + str(statement))
        try:
            # todo: for test, comment
            #time.sleep(2.999999)

            stype_statement = type(statement)

            if type(statement) == str:
                statement = statement.strip()
                if len(statement) < 1:
                    raise KeySqlConnectException(KeySqlExceptionMessage.EMPTY_STATEMENT)

                if statement[-1] == ';':
                    #all_statements = statement[:-1]  # statement[:-1].split(';')
                    statement = statement[:-1]
                if ';' in statement:
                    statements = self.extract_statements(statement)
                else:
                    statements = [statement]

                if len(statements) > 1 and len(args) != 0:
                    raise KeySqlException(-1, KeySqlExceptionMessage.PARAMETERIZED_ONE_QUERY_ONLY)
                elif len(statements) == 1 and len(args) > 0:
                    statements[0] = KeySqlParserQueryParameterized.process_input_query(statements[0], *args)

                # no arguments now
                #else:  # if len(statements) > 1:
                #return self.execute(statements, base_classes=base_classes, keep_str_res=keep_str_res,
                #                       check_inj=check_inj)

                    #return self.execute(statements[0], base_classes=base_classes, keep_str_res=keep_str_res,
                    #
                    #                     check_inj=check_inj)
                    #self._reply_header, reply = self._connector.send_resv(statements[0])
                #else:
                #    return self.execute(statements, base_classes=base_classes, keep_str_res=keep_str_res, check_inj=check_inj)

                    # return only the result of last statement
                    #for statement_i in all_statements:
                    #    reply = self.execute(statement_i, base_classes)
                    #return reply

                # statements = self.extract_statements(statement)
                all_replies = []
                for statement in statements:
                    short_message = self._connector.cut_message(str(statement))
                    #logging.debug('---> final_request1: ' + short_message)
                    logging.info('---> final_request1: ' + short_message)

                    self.cum_execute_time_in_sec = time.time()

                    self._reply_header, reply = self._connector.send_resv(statement, format)
                    self.cum_execute_time_in_sec = time.time() - self.cum_execute_time_in_sec

                    logging.debug('---> header: ' + str(self._reply_header))
                    short_reply = self._connector.cut_message(reply)
                    logging.debug('---> reply:  ' + short_reply)

                    #if type(reply) is str and len(reply) == 0:
                    #    return reply

                    reply1 = reply
                    #if len(reply1) == 0 and format.lower() in  ['json', 'keysql']:
                    #     reply1 = self._reply_header['instance_count']

                    if format.lower() == 'json':
                        #  json.loads(<res>) raises except if <res> is empty
                        if len(reply) > 0:
                            try:
                                short_reply = self._connector.cut_message(str(reply))
                                #print(short_reply, '============================================= short_reply1')
                                reply1 = json.loads(reply)

                                #TODO: delete later, experiment
                                """
                                #def customKeySqlObjectDecoder(catalogDict):
                                #    return namedtuple('XXX', catalogDict.keys())(*catalogDict.values())
    
                                # Parse JSON into an object with attributes corresponding to dict keys.
                                for i in range(1): #  item in reply1:
                                    #keysql_objects = [ json.loads(item, object_hook=customKeySqlObjectDecoder) for item in reply1]
                                    #print(type(keysql_objects), type(keysql_objects[0]),  keysql_objects, '------------------------------------ko')
                                    keysql_object = json.loads(reply, object_hook=customKeySqlObjectDecoder)
                                    print(type(keysql_object),  keysql_object, '------------------------------------ko experiment')
                                """

                            except KeySqlException as ex: # ((KeySqlExceptionMessage.WRONG_JSON_FORMAT) as ex:
                                short_reply = self._connector.cut_message(str(reply))
                                #print(short_reply, '=============================================short_reply2')
                                logging.error(ex.error_text)  # (short_reply)
                                raise KeySqlException(self._reply_header['exit_code'], short_reply)
                    #"""
                    elif format.lower() == 'keysql':
                        # if it is in pretty format
                        # reply = reply.replace('\n', '').replace('\t', '').replace(' ', '')
                        # reply = reply.replace('\t', '')
                        # if self._reply_header[]
                        reply1 = KeySqlParser.parse_select_reply(reply)

                    # only for select statement
                    elif format.lower() == 'packed':
                        #print("==== packed format === reply: ", reply)
                        parser_packed_format = KeySqlParserPackedFormat()
                        print(reply)
                        reply1 = parser_packed_format.process_select_reply(reply)
                    else:
                        logging.error(KeySqlExceptionMessage.FORMAT_KEYSQL_JSON)
                        raise KeySqlException(KeySqlExceptionMessage.FORMAT_KEYSQL_JSON)

                    #self._reply_header, reply = self._connector.send_resv(str.encode(statement, encoding="utf-8"))
                    if keep_str_res:
                        self.last_str_result = reply
                    short_reply1 = self._connector.cut_message(str(reply))

                    logging.debug('time: ' + str(self._reply_header["time"]) + '---> reply1: ' + short_reply1)
                    if self._reply_header['exit_code'] != 0:
                        short_reply = self._connector.cut_message(str(reply1))
                        raise KeySqlException(self._reply_header['exit_code'], short_reply)

                    if format.lower() == "keysql": #   != "packed":
                        reply = KeySqlReplyStatement(self._reply_header).get_keysql_statement().process_reply(reply1, self._reply_header)
                    else:
                        reply = reply1
                    all_replies.append(reply)

                return all_replies[-1]  # reply

            elif type(statement) == list:
                if len(statement) > 1 and len(args) > 0:
                    raise KeySqlException(-1, KeySqlExceptionMessage.PARAMETERIZED_ONE_QUERY_ONLY)
                if len(statement) == 1:  #  and len(args) > 0:
                    reply = self.execute(statement[0], *args, base_classes=base_classes, keep_str_res=keep_str_res,
                                         check_inj=check_inj)
                    return reply

                reply = ''
                # no arg here
                for st in statement:
                    reply = self.execute(st, base_classes=base_classes, keep_str_res=keep_str_res, check_inj=check_inj)
                return reply

            elif (KeySqlStatement in statement.__class__.__mro__) or \
                 (KeySqlShowStatement in statement.__class__.__mro__):   #  KeySqlStatement, KeySqlShowStatement
                reply = '***no_reply***'
                try:
                    final_request = statement.create_request()
                    #print(final_request, '--------------------------------final request')
                    short_request = self._connector.cut_message(str(final_request))
                    #logging.debug('---> final_request2: ' + short_request)
                    logging.info('---> final_request2: ' + short_request)

                    reply = self.execute(final_request, base_classes=base_classes, format=format, keep_str_res=keep_str_res, check_inj=check_inj)
                    #print(self.get_header_reply(), '------------------------------xxx')
                    logging.info('---> final_request2 executed with time: ' + str(self.get_header_reply()['time']))
                    return reply

                    """
                    if format == 'json':
                        if len(reply) > 0:
                            try:
                                reply1 = json.loads(reply)
                                print(reply1, '--- class json')
                            except KeySqlException(KeySqlExceptionMessage.WRONG_JSON_FORMAT) as ex:
                                raise KeySqlException(self._reply_header['exit_code'], reply)
                    elif format == 'keysql':
                        reply1 = KeySqlParser.parse_select_reply(reply)
                        pass
                    else:
                        raise KeySqlException(KeySqlExceptionMessage.FORMAT_KEYSQL_JSON)
    
                    self.cum_execute_time_in_sec = time.time()
                    self._reply_header, reply = self._connector.send_resv(final_request, format)
                    self.cum_execute_time_in_sec = time.time() - self.cum_execute_time_in_sec
    
                    logging.debug('time: ' + str(self._reply_header["time"]) + ' ---> reply: ' + reply + '\n')
                    if self._reply_header['exit_code'] == 0:
                        return statement.process_reply(reply, self._reply_header)
                    else:
                        logging.warning('--- not success: ', self._reply_header['exit_code'])
                        logging.warning('---', reply)
                        # print_debug('---', self._reply_header['exit_code'], reply)
                        if self._reply_header is not None:
                            raise KeySqlException(self._reply_header['exit_code'], reply)
                        else:
                            raise KeySqlException(-1, reply)
                    """
                except KeySqlException as kerr:
                    raise
                except ValueError as err:
                    logging.exception("message")
                    short_reply = self._connector.cut_message(str(reply))
                    raise KeySqlException(-1, short_reply)
            else:
                raise KeySqlException(-1, KeySqlExceptionMessage.STATEMENT_TYPE_NOT_IMPLEMENTED)
        except TimeoutError as ex:
            logging.exception(KeySqlExceptionMessage.QUERY_TIMEOUT_EXCEPTION)
            print(ex, 'timeout happened ==================================xx_start')
            self._connector.abort_query()
            print(ex, 'after abort query ==================================xx_end')
            raise ex


    @timeout(KeySqlConnectionConstants.QUERY_TIMEOUT_SEC)
    def execute_script(self, statement, *args, keep_str_res=False, check_inj=False, format='keysql',
                format_out='keysql'):
        """Execute query on KeySql server
        :param statement (string, list of class:
        :param keep_str_res  =  keep string response from server (how it is shown in KeySql studio)
                 To retrive string response:  self.get_last_str_result().   Default_value=False,
        :param check_inj=False,  - not used now
        :param format   - format of string to get from server (for get_last_str_result() function). Default_value='keysql'
        :param format_out - in what format to output result ('class"||json). Default_value='keysqclass' (python class).
                As of now, format_out = format
        :return: a list of classes||json    For one element the result is list of one element [elem]
        :raise: Exception if query can not be performed (no connection, wrong syntax, etc.)
        """

        # logging.info('start query: ' + str(statement))
        try:
            # todo: for test, comment
            # time.sleep(2.999999)

            if type(statement) != str:
                raise KeySqlConnectException(KeySqlExceptionMessage.EMPTY_STATEMENT)

            short_message = self._connector.cut_message(str(statement))
            # logging.debug('---> final_request1: ' + short_message)
            logging.info('---> final_request1: ' + short_message)

            self.cum_execute_time_in_sec = time.time()

            all_reply_pairs = self._connector.send_resv_script(statement, format)

            self.cum_execute_time_in_sec = time.time() - self.cum_execute_time_in_sec

            all_replies = []

            for reply_pair in all_reply_pairs:
                self._reply_header = reply_pair[0]
                logging.debug('---> header: ' + str(self._reply_header))
                reply_str = reply_pair[1]
                short_reply = self._connector.cut_message(reply_str)
                logging.debug('---> reply:  ' + short_reply)
                reply1 = reply_str
                if format == 'json':
                    #  json.loads(<res>) raises except if <res> is empty
                    if len(reply_str) > 0:
                        try:
                            short_reply = self._connector.cut_message(str(reply_str))
                            # print(short_reply, '============================================= short_reply1')
                            reply1 = json.loads(reply_str)

                        except KeySqlException as ex:  # ((KeySqlExceptionMessage.WRONG_JSON_FORMAT) as ex:
                            short_reply = self._connector.cut_message(str(reply_str))
                            # print(short_reply, '=============================================short_reply2')
                            logging.error(ex.error_text)  # (short_reply)
                            raise KeySqlException(self._reply_header['exit_code'], short_reply)
                # """
                elif format.lower() in ['keysql', 'json']:
                    # if it is in pretty format
                    # reply = reply.replace('\n', '').replace('\t', '').replace(' ', '')
                    # reply = reply.replace('\t', '')
                    # if self._reply_header[]
                    reply1 = KeySqlParser.parse_select_reply(reply_str)
                else:
                    logging.error(KeySqlExceptionMessage.FORMAT_KEYSQL_JSON)
                    raise KeySqlException(KeySqlExceptionMessage.FORMAT_KEYSQL_JSON)

            # statements = self.extract_statements(statement)

                if self._reply_header['exit_code'] != 0:
                    short_reply = self._connector.cut_message(str(reply))
                    raise KeySqlException(self._reply_header['exit_code'], short_reply)

                # print(self._reply_header, '---------------debug')
                reply = KeySqlReplyStatement(self._reply_header).get_keysql_statement().process_reply(reply1,
                                                                                                      self._reply_header)
                all_replies.append(reply)

            # print(all_replies, '-----------------------all_replies')
            return all_replies                 # TODO: delete was before  all_replies[-1]  # reply


        except TimeoutError as ex:
            logging.exception(KeySqlExceptionMessage.QUERY_TIMEOUT_EXCEPTION)
            print(ex, 'timeout happened ==================================xx_start')
            self._connector.abort_query()
            print(ex, 'after abort query ==================================xx_end')
            raise ex


    def extract_statements(self, statement: str):
        stmts = []
        start = 0
        started_value = False
        for i, c in enumerate(statement):
            if c == "'":
                if not started_value:
                    started_value = True
                else:
                    started_value = False
            elif c == ";" and not started_value:  # not started_value and i > start + 1:
                stmts.append(statement[start:i].strip())
                start = i + 1
        if started_value:
            raise KeySqlException(-1, KeySqlExceptionMessage.IMPROPER_END_OF_STATEMENT)

        if start + 1 < len(statement):
            stmts.append(statement[start:len(statement)])
        return stmts


if __name__ == "__main__":
    pass


