# /**********************************************************************************
# * Copyright (C) 2017-2020 by KEYARK, INC. - All Rights Reserved.
# * All information contained herein is proprietary and confidential to KEYARK, INC.
# * Any use, reproduction, or disclosure without the written
# * permission of KEYARK, INC. is prohibited.
# * Written by Natalia Belyaeva <nbelyaeva@keyark.com>
# ***********************************************************************************/

from keysql_client.keysql_store_statement import KeySqlCreateStoreStatement
from keysql_client.keysql_store_statement import KeySqlDropStoreStatement
from keysql_client.keysql_store_statement import KeySqlDropStoreIfExistsStatement
from keysql_client.keysql_store_statement import KeySqlAlterStoreNameStatement
from keysql_client.keysql_store_statement import KeySqlTruncateStoreStatement
from keysql_client.keysql_store_statement import KeySqlRenameStoreStatement
from keysql_client.keysql_insert import InsertIntoStoreInstance
from keysql_client.keysql_select import SelectStatement
from keysql_client.keysql_delete import DeleteStatement
from keysql_client.keysql_update import UpdateStatement

from keysql_client.keysql_statement import KeySqlStatement
from keysql_client.keysql_exception import KeySqlException, KeySqlExceptionMessage

class KeySqlStore:
    """Store
    All queries are executed by connection and return result.
    Examples:
    catalog_name, hosts_objs = conn.execute(store.show())
    # to compare catalog and store in keysql format
    # catalog
    # {CATALOG_INFO:{SCHEMA_NAME:'TEST3',CATALOG_NAME:'XYZ',STORE_NAMES:NULL,
    # OWNER_NAME:'test3',
    # KEYOBJECT_DEFINITIONS:{KEYOBJECT_DEFINITION:{KEYOBJECT_NAME:'NN',KEYOBJECT_TYPE:'NUMBER',COMPONENT_KEYOBJECT_NAMES:NULL},
    # KEYOBJECT_DEFINITION:{KEYOBJECT_NAME:'CC',KEYOBJECT_TYPE:'CHAR',COMPONENT_KEYOBJECT_NAMES:NULL},
    # KEYOBJECT_DEFINITION:{KEYOBJECT_NAME:'CMPND',KEYOBJECT_TYPE:'COMPOSITION',COMPONENT_KEYOBJECT_NAMES:{KEYOBJECT_NAME:'NN',KEYOBJECT_NAME:'CC'}}}}}
    # store
    # {STORE_INFO:{SCHEMA_NAME:'TEST3',CATALOG_NAME:'PEOPLE',STORE_NAME:'PEOPLE_STORE',
    # OWNER_NAME:'test3',
    # HOSTS:{KEYOBJECT_DEFINITION:{KEYOBJECT_NAME:'PERSON',KEYOBJECT_TYPE:'COMPOSITION',
    # COMPONENT_KEYOBJECT_NAMES:{KEYOBJECT_NAME:'EMAILS',KEYOBJECT_NAME:'FIRST_NAME',KEYOBJECT_NAME:'LAST_NAME',KEYOBJECT_NAME:'YOB',KEYOBJECT_NAME:'ADDRESS'}}},
    # CONSTRAINTS:NULL}}
    """

    def __init__(self, name: str, catalog_name: str, schema_name="", owner_name="",
                 hosts=[], constraints=[]):
        """Create instance of KeySqlStore class
        :param name (str): store name
        Example:
        store = KeySqlStore('store_name')
        """
        self.__name__ = name
        self._catalog_name = catalog_name
        self._owner_name = owner_name
        self._schema_name = schema_name
        self._hosts = hosts if hosts else []
        self._constraints = constraints if constraints else []


    def store_name(self) -> str:
        """Store name
        :return (str): store name
        """
        return self.__name__

    def schema_name(self):
        """
        :return: return (str): schema name
        """
        return self._schema_name


    def catalog_name(self):
        """
        :return: return (str): catalog name
        """
        return self._catalog_name


    def owner_name(self):
        """
        :return:  return (str): Owner name
        """
        return self._owner_name


    def hosts(self):
        """
        :return: return (list): hosts
        """
        return self._hosts


    def constraints(self):
        """
        :return: return (list): constraints
        """
        return self._constraints


    def create(self, catalog_name: str) -> KeySqlStatement:
        """Create store <store_name> for catalog <catalog_name>
        :param (str) catalog_name: non-empty unique catalog name
        :return (KeySqlCreateStoreStatement) instance of
        Example:
        KeySqlCreateStoreStatement('store_name', 'catalog_name')
        """
        # TODO: get rid of catalog_name
        #return KeySqlCreateStoreStatement(self.__name__, catalog_name)
        return KeySqlCreateStoreStatement(self.__name__, self._catalog_name)

        #return KeySqlCreateStoreStatement(self.__name__, catalog_name, schema_name="", owner_name="", host_names=[], hosts=[], component_names=[], constraints=[])

    def get_name(self) -> str:
        """
        :return: str - Keyobject name
        """
        return self.__name__

    def serialize(self):
        return self.store_name()

    def drop(self) -> KeySqlStatement:
        """Drop Store <store_name>
        :return (KeySqlDropStoreStatement): instance of class KeySqlDropStoreStatement
        Example:
        store.drop() , where store is instance of class KeySqlStore
        """
        return KeySqlDropStoreStatement(self.__name__)

    def drop_if_exists(self) -> KeySqlStatement:
        """Drop Store <store_name>
        :return (KeySqlDropStoreStatement): instance of class KeySqlDropStoreStatement
        Example:
        store.drop() , where store is instance of class KeySqlStore
        """
        return KeySqlDropStoreIfExistsStatement(self.__name__)

    def rename(self, name : str, schema=None) -> KeySqlStatement:
        old_name = self.__name__
        self.__name__ = name
        return KeySqlRenameStoreStatement(old_name, self.__name__, schema=schema)

    def alter(self, name: str, catalog_name: str) -> KeySqlStatement:
        """Alter store name
        :param name (str): new store name
        :param catalog_name (str): catalog name
        :return (KeySqlAlterStoreNameStatement): instance of KeySqlAlterStoreNameStatement
        Example:
        store.alter('new_store_name', 'catalog_name')
        """
        old_name = self.__name__
        self.__name__ = name
        return KeySqlAlterStoreNameStatement(old_name, self.__name__, catalog_name)

    def insert(self, instances) -> KeySqlStatement:
        """Insert instances into the store
        :param instances (instance or list of instances, where instance can be a dictionary or a class):
        :return (InsertIntoStoreInstance): instance of InsertIntoStoreInstance
        Example:
        store.insert([
            Person(last_name='Johns', yob=1999, emails=[Email(email_address='shopping@johns.com'),
                                                        Email(email_address='amazon.shopping@johns.com')]),
            {"Person":
                {"first_name": 'John',
                "last_name": 'Smith',
                "yob": 1997,
                "address": {"country": "US", "city": "San francisco"},
                "emails": [Email(email_address = "smith@work.org", email_type="work"),
                Email(email_address = "John@smith.org", email_type="home")]
                }
            }])
        """
        return InsertIntoStoreInstance(self.__name__, instances)

    def select(self, *args) -> KeySqlStatement:
        """Select
        :param args (str, any class): variable number
        :return (KeySqlStatement):
        Example:
        store.select("last_name","email")
        store.select(Person())  where Person is class Person created by an user
        """
        return SelectStatement()._from(self.__name__).select(*args)

    def delete(self, where_str: str = None) -> None:
        """Delete
        :param where_str (str): where condition as string
        :return: None
        Example:
        store.delete().where("last_name = 'Johns'")
        """
        if where_str:
            return DeleteStatement()._from(self.__name__).where(where_str)
        return DeleteStatement()._from(self.__name__)

    def update(self) -> None:
        """Update store
        :return: None
        Example:
        store.update()
        """
        return UpdateStatement(self.__name__)

    def truncate(self) -> None:
        """Truncate store  - delete all keyobjects and instances => result: empty store
        :return: None
        Example:
        store.truncate()
        """
        return KeySqlTruncateStoreStatement(self._schema_name, self.__name__)

